//
// Grass rendering example 
//
// Author: Alexey V. Boreskov <steps3d@gmail.com>, <steps3d@narod.ru>
//

#include	"GlutCameraWindow.h"
#include	"Program.h"
#include	"BasicMesh.h"
#include	"Texture.h"
#include	"randUtils.h"


class	RainWindow : public GlutCameraWindow
{
	Program		    rainProgram;
	Texture 	    precipitationMap;			// for terrain
	Texture		    noiseMap;			// 3D-texture with noise
	VertexArray	    vao;				// array with bindings for grass
	VertexBuffer	buf;
	int		        gridSize  = 7;
	float		    threshold = 0.05;
	int		        nx        = 100;
	int		        ny        = 100;
	const float		zNear     = 0.01f;
	const float		zFar      = 500.0f;

	struct	Vertex
	{
		glm::vec4	pos;
		glm::vec2	tex;
	};

public:
	RainWindow () : GlutCameraWindow ( 100, 100, 900, 900, "Rain streaks with instancing" ) 
	{
		if ( !rainProgram.loadProgram ( "rain.glsl" ) )
			exit ( "Error loading shader: %s\n", rainProgram.getLog ().c_str () );
		
		rainProgram.bind       ();
		rainProgram.setTexture ( "noiseMap",         0 );
		rainProgram.setTexture ( "precipitationMap", 1 );
		rainProgram.unbind     ();
		
		if ( !noiseMap.load3D ( "../../Textures/noise-3D.dds" ) )
			exit ( "Error loading noise texture\n" );

		if ( !precipitationMap.load2D ( "Precipitation.png" ) )
			exit ( "Error loading  Precipitation  texture\n" );

		noiseMap.bind         ( 0 );
		precipitationMap.bind ( 1 );
		camera.setZ           ( zNear, zFar  );
		prepareVertices       ();
	}
	
	void redisplay () override
	{
		glClear     ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
		glEnable    ( GL_BLEND );
		glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

		rainProgram.bind             ();
		rainProgram.setUniformMatrix ( "mv",         getModelView  () );
		rainProgram.setUniformMatrix ( "proj",       getProjection () );
		rainProgram.setUniformInt    ( "gridSize",   gridSize );
		rainProgram.setUniformVector ( "screenSize", glm::vec2 ( getWidth (), getHeight () ) );
		rainProgram.setUniformVector ( "eye",        camera.getPos () );
		rainProgram.setUniformInt    ( "nx",         nx );
		rainProgram.setUniformInt    ( "ny",         ny );
		rainProgram.setUniformFloat  ( "threshold",  threshold );
		rainProgram.setUniformFloat  ( "time",       getTime () );

		vao.bind ();
		glDrawArraysInstanced ( GL_POINTS, 0, nx*ny, 27 );		// 100^2 vertices, 27 instances
		vao.unbind ();
	}
	
	void	keyTyped ( unsigned char key, int modifiers, int x, int y ) override
	{
		if ( key == '+' && threshold < 1 )
			threshold += 0.01;
		else if ( key == '-' && threshold > 0 )
			threshold -= 0.01;

		GlutCameraWindow :: keyTyped ( key, modifiers, x, y );
	}

	void	prepareVertices ()
	{
		std::vector<Vertex>	vertices ( nx * ny );
		Vertex				v;
		int					c = 0;

		for ( int i = 0; i < nx; i++ )
			for ( int j = 0; j < ny; j++ )
			{
                v.pos   = glm::vec4 ( i / (float) nx, 0.0f, j / (float) ny, randUniform () );
                v.tex   = glm::vec2 ( v.pos.x, v.pos.y );

				vertices [c++] = v;
			}

		vao.create  ();
		vao.bind    ();
		buf.create  ();
		buf.bind    ( GL_ARRAY_BUFFER );
		buf.setData ( vertices, GL_STATIC_DRAW );

		glVertexAttribPointer ( 0, 			// index
							4, 			    // number of values per vertex
							GL_FLOAT, 
							GL_FALSE,
							sizeof(Vertex),	// stride (offset to next vertex data)
							(const GLvoid*) 0 );
		
		glEnableVertexAttribArray ( 0 );
		glVertexAttribPointer ( 1, 			// index
							2, 			    // number of values per vertex
							GL_FLOAT, 
							GL_FALSE,
							sizeof(Vertex),	// stride (offset to next vertex data)
							(const GLvoid*) (4*sizeof(float)) );
		
		glEnableVertexAttribArray ( 1 );

		vao.unbind  ();
	}

};

int main ( int argc, char * argv [] )
{
	GlutWindow::init ( argc, argv );
	
	RainWindow	win;
	
	return GlutWindow::run ();
}
